import argparse
import logging
import sys
from datetime import datetime
from typing import Sequence

from dateutil.tz import tzutc
from kubernetes.client import V1Pod

from paasta_tools.kubernetes_tools import get_all_pods
from paasta_tools.kubernetes_tools import get_pod_condition
from paasta_tools.kubernetes_tools import is_pod_completed
from paasta_tools.kubernetes_tools import KubeClient

log = logging.getLogger(__name__)


def parse_args():
    parser = argparse.ArgumentParser(
        description="Terminates completed pods based on namespace and minutes since completion"
    )
    parser.add_argument(
        "-n",
        "--namespace",
        help="Namespace of the pods to terminate from",
        required=True,
    )
    parser.add_argument(
        "-m",
        "--minutes",
        help="Minutes since the pods' completion. Terminates pods based on time since completion.",
        required=True,
        type=int,
    )
    parser.add_argument(
        "-e",
        "--error-minutes",
        help="Minutes since the pod encountered an error. Terminates pods based on time since failure.",
        # this can't be required until we've rolled this out everywhere AND have updated all the callsites
        required=False,
        type=int,
    )
    parser.add_argument(
        "-p",
        "--pending-minutes",
        help="Minutes since the pod was scheduled. Terminates pods whose phase is Pending based on time since scheduled. Including pod status Pending/ContainerCreating/Terminating.",
        required=False,
        type=int,
    )
    parser.add_argument(
        "--dry-run",
        dest="dry_run",
        action="store_true",
        help="Print pods to be terminated, instead of terminating them",
    )
    parser.add_argument(
        "-v", "--verbose", dest="verbose", action="store_true", default=False
    )
    args = parser.parse_args()
    return args


def setup_logging(verbose):
    level = logging.DEBUG if verbose else logging.WARNING
    logging.basicConfig(level=level)
    # Remove pod metadata logs
    logging.getLogger("kubernetes.client.rest").setLevel(logging.ERROR)


def __condition_transition_longer_than_threshold(
    pod: V1Pod, condition: str, threshold: int
) -> bool:
    time_finished = get_pod_condition(pod, condition).last_transition_time
    time_now = datetime.now(tzutc())

    # convert total seconds since completion to minutes
    since_minutes = (time_now - time_finished).total_seconds() / 60

    return since_minutes > threshold


def _completed_longer_than_threshold(pod: V1Pod, threshold: int) -> bool:
    return __condition_transition_longer_than_threshold(
        pod, "ContainersReady", threshold
    )


def _scheduled_longer_than_threshold(pod: V1Pod, threshold: int) -> bool:
    return __condition_transition_longer_than_threshold(pod, "PodScheduled", threshold)


def terminate_pods(pods: Sequence[V1Pod], kube_client) -> tuple:
    successes = []
    errors = []

    for pod in pods:
        try:
            kube_client.core.delete_namespaced_pod(
                name=pod.metadata.name,
                namespace=pod.metadata.namespace,
                grace_period_seconds=0,
                propagation_policy="Background",
            )
            successes.append(pod.metadata.name)
        except Exception as e:
            errors.append((pod.metadata.name, e))

    return (successes, errors)


def main():
    args = parse_args()
    setup_logging(args.verbose)

    kube_client = KubeClient()
    pods = get_all_pods(kube_client, args.namespace)

    allowed_uptime_minutes = args.minutes
    allowed_error_minutes = args.error_minutes
    allowed_pending_minues = args.pending_minutes

    completed_pods = []
    errored_pods = []
    pending_pods = []

    for pod in pods:
        if is_pod_completed(pod) and _completed_longer_than_threshold(
            pod, allowed_uptime_minutes
        ):
            completed_pods.append(pod)
        elif (
            # this is currently optional
            allowed_error_minutes is not None
            # there's no direct way to get what type of "bad" state these Pods ended up
            # (kubectl looks at phase and then container statuses to give something descriptive)
            # but, in the end, we really just care that a Pod is in a Failed phase
            and pod.status.phase == "Failed"
        ):
            try:
                # and that said Pod has been around for a while (generally longer than we'd leave
                # Pods that exited sucessfully)
                # NOTE: we do this in a try-except since we're intermittently seeing pods in an error
                # state without a PodScheduled condition (even though that should be impossible)
                # this is not ideal, but its fine to skip these since this isn't a critical process
                if _scheduled_longer_than_threshold(pod, allowed_error_minutes):
                    errored_pods.append(pod)
            except AttributeError:
                log.exception(
                    f"Unable to check {pod.metadata.name}'s schedule time. Pod status: {pod.status}.'"
                )
        elif (
            # this is currently optional
            allowed_pending_minues is not None
            and pod.status.phase == "Pending"
        ):
            try:
                if _scheduled_longer_than_threshold(pod, allowed_pending_minues):
                    pending_pods.append(pod)
            except AttributeError:
                log.exception(
                    f"Unable to check {pod.metadata.name}'s schedule time. Pod status: {pod.status}.'"
                )

    if not (completed_pods or errored_pods or pending_pods):
        log.debug("No pods to terminate.")
        sys.exit(0)

    if args.dry_run:
        log.debug(
            "Dry run would have terminated the following completed pods:\n "
            + "\n ".join([pod.metadata.name for pod in completed_pods])
        )
        log.debug(
            "Dry run would have terminated the following errored pods:\n "
            + "\n ".join([pod.metadata.name for pod in errored_pods])
        )
        log.debug(
            "Dry run would have terminated the following pending pods:\n "
            + "\n ".join([pod.metadata.name for pod in pending_pods])
        )
        sys.exit(0)

    completed_successes, completed_errors = terminate_pods(completed_pods, kube_client)
    errored_successes, errored_errors = terminate_pods(errored_pods, kube_client)
    pending_successes, pending_errors = terminate_pods(pending_pods, kube_client)

    successes = {
        "completed": completed_successes,
        "errored": errored_successes,
        "pending": pending_successes,
    }
    errors = {
        "completed": completed_errors,
        "errored": errored_errors,
        "pending": pending_errors,
    }

    for typ, pod_names in successes.items():
        if pod_names:
            log.debug(
                f"Successfully terminated the following {typ} pods:\n"
                + "\n ".join(pod_names)
            )

    # we've only really seen this fail recently due to the k8s API being flaky and returning
    # 404s for Pods that its returning to us when we get all Pods, so we just print the error
    # here for now and don't exit with a non-zero exit code since, again, this isn't a critical
    # process
    for typ, pod_names_and_errors in errors.items():
        if pod_names_and_errors:
            log.error(
                f"Failed to terminate the following {typ} pods:\n"
                + "\n  ".join(
                    f"{pod_name}: {error}" for pod_name, error in pod_names_and_errors
                )
            )


if __name__ == "__main__":
    main()



import argparse
import logging
import os.path
import time
from collections import defaultdict

from inotify.adapters import Inotify
from inotify.constants import IN_MODIFY
from inotify.constants import IN_MOVED_TO

from paasta_tools import firewall
from paasta_tools.cli.utils import get_instance_config
from paasta_tools.utils import DEFAULT_SOA_DIR
from paasta_tools.utils import load_system_paasta_config
from paasta_tools.utils import TimeoutError

log = logging.getLogger(__name__)

DEFAULT_UPDATE_SECS = 5


def parse_args(argv):
    parser = argparse.ArgumentParser(
        description="Monitor synapse changes and update service firewall rules"
    )
    parser.add_argument(
        "-d",
        "--soa-dir",
        dest="soa_dir",
        metavar="soa_dir",
        default=DEFAULT_SOA_DIR,
        help="define a different soa config directory (default %(default)s)",
    )
    parser.add_argument(
        "--synapse-service-dir",
        dest="synapse_service_dir",
        default=firewall.DEFAULT_SYNAPSE_SERVICE_DIR,
        help="Path to synapse service dir (default %(default)s)",
    )
    parser.add_argument("-v", "--verbose", dest="verbose", action="store_true")

    subparsers = parser.add_subparsers(
        help="mode to run firewall update in", dest="mode"
    )
    subparsers.required = True

    daemon_parser = subparsers.add_parser(
        "daemon",
        description=(
            "Run a daemon which watches updates to synapse backends and updates iptables rules."
        ),
    )
    daemon_parser.add_argument(
        "-u",
        "--update-secs",
        dest="update_secs",
        default=DEFAULT_UPDATE_SECS,
        type=int,
        help="Poll for new containers every N secs (default %(default)s)",
    )

    subparsers.add_parser(
        "cron",
        description=(
            "Do a one-time update of iptables rules to match the current running services."
        ),
    )

    args = parser.parse_args(argv)
    return args


def setup_logging(verbose):
    level = logging.DEBUG if verbose else logging.WARNING
    logging.basicConfig(level=level)


def run_daemon(args):
    # Main loop waiting on inotify file events
    inotify = Inotify(block_duration_s=1)  # event_gen blocks for 1 second
    inotify.add_watch(args.synapse_service_dir.encode(), IN_MOVED_TO | IN_MODIFY)
    services_by_dependencies_time = 0

    for event in inotify.event_gen():  # blocks for only up to 1 second at a time
        if services_by_dependencies_time + args.update_secs < time.time():
            services_by_dependencies = (
                smartstack_dependencies_of_running_firewalled_services(
                    soa_dir=args.soa_dir
                )
            )
            services_by_dependencies_time = time.time()

        if event is None:
            continue

        process_inotify_event(
            event, services_by_dependencies, args.soa_dir, args.synapse_service_dir
        )


def run_cron(args):
    with firewall.firewall_flock():
        firewall.general_update(args.soa_dir, args.synapse_service_dir)


def process_inotify_event(
    event, services_by_dependencies, soa_dir, synapse_service_dir
):
    filename = event[3].decode()
    log.debug(f"process_inotify_event on {filename}")

    service_instance, suffix = os.path.splitext(filename)
    if suffix != ".json":
        return

    services_to_update = services_by_dependencies.get(service_instance, ())
    if not services_to_update:
        return

    # filter active_service_groups() down to just the names in services_to_update
    service_groups = {
        service_group: macs
        for service_group, macs in firewall.active_service_groups().items()
        if service_group in services_to_update
    }

    try:
        with firewall.firewall_flock():
            firewall.ensure_service_chains(service_groups, soa_dir, synapse_service_dir)

        for service_to_update in services_to_update:
            log.debug(f"Updated {service_to_update}")
    except TimeoutError as e:
        log.error(
            "Unable to update firewalls for {} because time-out obtaining flock: {}".format(
                service_groups.keys(), e
            )
        )


def smartstack_dependencies_of_running_firewalled_services(soa_dir=DEFAULT_SOA_DIR):
    dependencies_to_services = defaultdict(set)
    for service, instance, _, _ in firewall.services_running_here():
        config = get_instance_config(
            service=service,
            instance=instance,
            cluster=load_system_paasta_config().get_cluster(),
            load_deployments=False,
            soa_dir=soa_dir,
        )
        inbound_firewall = config.get_inbound_firewall()
        outbound_firewall = config.get_outbound_firewall()
        if not inbound_firewall and not outbound_firewall:
            continue

        dependencies = config.get_dependencies() or ()

        smartstack_dependencies = [
            d["smartstack"] for d in dependencies if d.get("smartstack")
        ]
        for smartstack_dependency in smartstack_dependencies:
            # TODO: filter down to only services that have no proxy_port
            dependencies_to_services[smartstack_dependency].add(
                firewall.ServiceGroup(service, instance)
            )

    return dependencies_to_services


def main(argv=None):
    args = parse_args(argv)
    setup_logging(args.verbose)
    {"daemon": run_daemon, "cron": run_cron}[args.mode](args)
